#    sizzlews
#    Copyright (C) 2020 Dmitry Berezovsky
#    The MIT License (MIT)
#    
#    Permission is hereby granted, free of charge, to any person obtaining
#    a copy of this software and associated documentation files
#    (the "Software"), to deal in the Software without restriction,
#    including without limitation the rights to use, copy, modify, merge,
#    publish, distribute, sublicense, and/or sell copies of the Software,
#    and to permit persons to whom the Software is furnished to do so,
#    subject to the following conditions:
#    
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#    
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
#    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
#    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import json
import logging
from abc import ABCMeta, abstractmethod
from typing import Union, Any, Dict, List, Type, Optional

import pydantic

RESPONSE_FIELD_RESULT = 'result'

RESPONSE_FIELD_EROR = 'error'

RESPONSE_FIELD_JSONRPC = 'jsonrpc'

InvocationParams = Union[Dict[str, Any], List[Any]]
InvocationResultType = Type[Union[pydantic.BaseModel, List[pydantic.BaseModel]]]
InvocationResult = Union[pydantic.BaseModel, List[pydantic.BaseModel]]


class JsonRpcRequest(object):
    def __init__(self, method, params: InvocationParams = None) -> None:
        super().__init__()
        self.method = method
        self.paams = params
        self.id = 0

    def to_dict(self) -> dict:
        return {
            "method": self.method,
            "params": self.paams,
            "jsonrpc": "2.0",
            "id": self.id
        }

    def to_json(self) -> str:
        return json.dumps(self.to_dict())


class RPCSourceError(Exception):

    def __init__(self, msg: str, type: str, args: List[Any]) -> None:
        super().__init__(msg)
        self.type: str = type
        self.msg: str = msg
        self.args: List[Any] = args


class RPCInvocationError(Exception):

    def __init__(self, msg: str, code: int, data: Dict) -> None:
        super().__init__(msg)
        self.msg: str = msg
        self.code: int = code
        self.data: Dict = data
        self.error: Optional[RPCSourceError] = None

        # If there is data field - try to parse it
        if data is not None:
            self.error = RPCSourceError(
                msg=data.get('message', 'Unknown error'),
                type=data.get('type', 'UNKNOWN'),
                args=data.get('args', []),
            )

    @classmethod
    def from_rpc_dict(cls, rpc_dict: dict) -> 'RPCInvocationError':
        return cls(rpc_dict['message'], rpc_dict['code'], rpc_dict.get('data', None))

    def __repr__(self) -> str:
        if self.error is None:
            return self.msg
        return 'RPC Invocation error {}: {}. Cause: {}: {}'.format(str(self.code), self.msg, self.error.type,
                                                                   self.error.msg)

    def __str__(self):
        return self.__repr__()


class SizzleWsBaseClient(metaclass=ABCMeta):
    def _params_to_rq(self, method_name, args, kwargs) -> JsonRpcRequest:
        params = []
        if args and kwargs:
            raise ValueError("You can't mix named and list arguments")
        if args:
            params = args
        if kwargs:
            params = kwargs
        rq = JsonRpcRequest(method_name, params)
        return rq

    @staticmethod
    def _parse_rpc_response(response: Dict[str, Any],
                            expected_response_type: InvocationResultType = None) -> InvocationResult:
        if RESPONSE_FIELD_JSONRPC not in response or response[RESPONSE_FIELD_JSONRPC] != '2.0':
            raise ValueError('Unknown response format. Expected result should be in JSONRPC 2.0 format')
        if RESPONSE_FIELD_EROR in response:
            raise RPCInvocationError.from_rpc_dict(response[RESPONSE_FIELD_EROR])
        elif RESPONSE_FIELD_RESULT in response:
            if expected_response_type is None:
                return response[RESPONSE_FIELD_RESULT]
            else:
                if isinstance(response[RESPONSE_FIELD_RESULT], list):
                    return list(map(lambda x: expected_response_type.parse_obj(x), response[RESPONSE_FIELD_RESULT]))
                else:
                    return expected_response_type.parse_obj(response[RESPONSE_FIELD_RESULT])


class SizzleWsClient(SizzleWsBaseClient, metaclass=ABCMeta):

    def __init__(self) -> None:
        self._logger = logging.getLogger("SizzleWsClient")

    def init(self):
        pass

    def close(self):
        pass

    def invoke(self, method_name, *args, expected_response_type: InvocationResultType = None, **kwargs):
        rq = self._params_to_rq(method_name, args, kwargs)
        return self._invoke_request(rq, expected_response_type)

    @abstractmethod
    def _invoke_request(self, rq: JsonRpcRequest, expected_response_type: InvocationResultType = None):
        pass

    def __enter__(self):
        self.init()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            self.close()
            return True
        except Exception as e:
            self._logger.exception("Error while closing SizzleWsClient", e)


class SizzleWsAsyncClient(SizzleWsBaseClient, metaclass=ABCMeta):

    def __init__(self) -> None:
        super().__init__()
        self._logger = logging.getLogger("SizzleWsAsyncClient")

    async def init(self):
        pass

    async def close(self):
        pass

    @abstractmethod
    async def _invoke_request(self, rq: JsonRpcRequest, expected_response_type: InvocationResultType = None):
        pass

    async def async_invoke(self, method_name, *args, expected_response_type: InvocationResultType = None, **kwargs):
        rq = self._params_to_rq(method_name, args, kwargs)
        return await self._invoke_request(rq, expected_response_type)

    def __enter__(self):
        raise TypeError("Use async with instead")

    async def __aenter__(self):
        await self.init()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        try:
            await self.close()
            return True
        except Exception as e:
            self._logger.exception("Error while closing SizzleWsClient", e)



#    sizzlews
#    Copyright (C) 2020 Dmitry Berezovsky
#    The MIT License (MIT)
#    
#    Permission is hereby granted, free of charge, to any person obtaining
#    a copy of this software and associated documentation files
#    (the "Software"), to deal in the Software without restriction,
#    including without limitation the rights to use, copy, modify, merge,
#    publish, distribute, sublicense, and/or sell copies of the Software,
#    and to permit persons to whom the Software is furnished to do so,
#    subject to the following conditions:
#    
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#    
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
#    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
#    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import asyncio
from typing import List

import pydantic

from sizzlews.client.aiohttp import SizzleWsAIOClient
from sizzlews.client.common import RPCInvocationError


class MyDTO(pydantic.BaseModel):
    field1: int
    field2: str


class MyTestApiClient(SizzleWsAIOClient):

    async def some_method(self, a: int, b: int):
        return await self.async_invoke('api.some_method', a, b)

    async def divide_by_zero(self, a: int):
        return await self.async_invoke('api.divide_by_zero', a)

    async def my_dto_method(self):
        return await self.async_invoke('api.my_dto_method', expected_response_type=MyDTO)

    async def returns_list_of_dtos(self) -> List[MyDTO]:
        return await self.async_invoke('api.returns_list_of_dtos', expected_response_type=MyDTO)


client = MyTestApiClient('http://localhost:8888/rpc')


async def main():
    async with client:
        print(await client.some_method(1, 2))
        try:
            print(await client.divide_by_zero(1))
        except RPCInvocationError as e:
            print("Error: " + e.msg)
        try:
            print(await client.my_dto_method())
        except Exception as e:
            print(e)
        try:
            print("List of DTOs")
            print(await client.returns_list_of_dtos())
        except Exception as e:
            print(e)


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())



#    sizzlews
#    Copyright (C) 2020 Dmitry Berezovsky
#    The MIT License (MIT)
#    
#    Permission is hereby granted, free of charge, to any person obtaining
#    a copy of this software and associated documentation files
#    (the "Software"), to deal in the Software without restriction,
#    including without limitation the rights to use, copy, modify, merge,
#    publish, distribute, sublicense, and/or sell copies of the Software,
#    and to permit persons to whom the Software is furnished to do so,
#    subject to the following conditions:
#    
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#    
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
#    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
#    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import asyncio

import websockets
from websockets import WebSocketServerProtocol

from sizzlews.server.common import SizzleWSHandler, BiDirectionalSizzleWSHandler, BidirectionalSizzleWSSession


def WebsocketsSizzleWSHandler(api_handler: BiDirectionalSizzleWSHandler):
    async def ws_handler(ws: WebSocketServerProtocol, path: str):
        if path != '/rpc':
            await ws.close(404, "not found")
        session = BidirectionalSizzleWSSession(id(ws), connection=ws)
        api_handler.register_session(session)
        async for message in ws:
            await ws.send(api_handler.handle(message, session))

    return ws_handler


def bootstrap_websockets_rpc_application(api_handler: SizzleWSHandler, port=8888, url_path='/',
                                         host: str = '0.0.0.0', **kwargs):
    start_ws_server = websockets.serve(WebsocketsSizzleWSHandler(api_handler), host, port, **kwargs)
    asyncio.get_event_loop().run_until_complete(start_ws_server)



#    sizzlews
#    Copyright (C) 2020 Dmitry Berezovsky
#    The MIT License (MIT)
#    
#    Permission is hereby granted, free of charge, to any person obtaining
#    a copy of this software and associated documentation files
#    (the "Software"), to deal in the Software without restriction,
#    including without limitation the rights to use, copy, modify, merge,
#    publish, distribute, sublicense, and/or sell copies of the Software,
#    and to permit persons to whom the Software is furnished to do so,
#    subject to the following conditions:
#    
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#    
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
#    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
#    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import asyncio

from sizzlews.server.annotation import rpc_method
from sizzlews.server.common import SizzleWSHandler, MethodDiscoveryMixin
from sizzlews.server.websockets import bootstrap_websockets_rpc_application


class MyApi(MethodDiscoveryMixin, SizzleWSHandler):
    METHOD_PREFXIX = "api."

    @rpc_method
    def some_method(self, a: int, b):
        return a + b


if __name__ == "__main__":
    bootstrap_websockets_rpc_application(MyApi(), 5555, '/wsrpc')
    asyncio.get_event_loop().run_forever()



#    sizzlews
#    Copyright (C) 2020 Dmitry Berezovsky
#    The MIT License (MIT)
#    
#    Permission is hereby granted, free of charge, to any person obtaining
#    a copy of this software and associated documentation files
#    (the "Software"), to deal in the Software without restriction,
#    including without limitation the rights to use, copy, modify, merge,
#    publish, distribute, sublicense, and/or sell copies of the Software,
#    and to permit persons to whom the Software is furnished to do so,
#    subject to the following conditions:
#    
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#    
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
#    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
#    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import aiohttp

from sizzlews.client.common import JsonRpcRequest, InvocationResultType, SizzleWsAsyncClient


class SizzleWsAIOClient(SizzleWsAsyncClient):

    def __init__(self, endpoint: str = None) -> None:
        super().__init__()
        self.endpoint = endpoint
        self._session: aiohttp.ClientSession = None

    async def init(self):
        self._session = aiohttp.ClientSession()

    async def close(self):
        if self._session:
            await self._session.close()

    async def _invoke_request(self, rq: JsonRpcRequest, expected_response_type: InvocationResultType = None):
        if self._session is None:
            raise ValueError('Client is not initialized. Call init before use')
        response = await self._session.post(self.endpoint, json=rq.to_dict())
        return self._parse_rpc_response(await response.json(), expected_response_type)



#    sizzlews
#    Copyright (C) 2020 Dmitry Berezovsky
#    The MIT License (MIT)
#    
#    Permission is hereby granted, free of charge, to any person obtaining
#    a copy of this software and associated documentation files
#    (the "Software"), to deal in the Software without restriction,
#    including without limitation the rights to use, copy, modify, merge,
#    publish, distribute, sublicense, and/or sell copies of the Software,
#    and to permit persons to whom the Software is furnished to do so,
#    subject to the following conditions:
#    
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#    
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
#    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
#    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import inspect
import json
import logging

import pydantic
from jsonrpc.exceptions import (
    JSONRPCInvalidParams,
    JSONRPCInvalidRequest,
    JSONRPCInvalidRequestException,
    JSONRPCMethodNotFound,
    JSONRPCParseError,
    JSONRPCServerError,
    JSONRPCDispatchException,
)
from jsonrpc.jsonrpc import JSONRPCRequest
from jsonrpc.jsonrpc1 import JSONRPC10Response
from jsonrpc.jsonrpc2 import (
    JSONRPC20BatchRequest,
    JSONRPC20BatchResponse,
    JSONRPC20Response,
)
from jsonrpc.utils import is_invalid_params

logger = logging.getLogger(__name__)


class JSONRPCResponseAsyncManager(object):
    """
    JSON-RPC response manager.

    Method brings syntactic sugar into library. Given dispatcher it handles
    request (both single and batch) and handles errors.
    Request could be handled in parallel, it is server responsibility.

    :param str request_str: json string. Will be converted into
        JSONRPC20Request, JSONRPC20BatchRequest or JSONRPC10Request

    :param dict dispather: dict<function_name:function>.

    """

    RESPONSE_CLASS_MAP = {
        "1.0": JSONRPC10Response,
        "2.0": JSONRPC20Response,
    }

    @classmethod
    async def handle(cls, request_str, dispatcher):
        if isinstance(request_str, bytes):
            request_str = request_str.decode("utf-8")

        try:
            data = json.loads(request_str)
        except (TypeError, ValueError):
            return JSONRPC20Response(error=JSONRPCParseError()._data)

        try:
            request = JSONRPCRequest.from_data(data)
        except JSONRPCInvalidRequestException:
            return JSONRPC20Response(error=JSONRPCInvalidRequest()._data)

        return await cls.handle_request(request, dispatcher)

    @classmethod
    async def handle_request(cls, request, dispatcher):
        """
        Handle request data.

        At this moment request has correct jsonrpc format.

        :param dict request: data parsed from request_str.
        :param jsonrpc.dispatcher.Dispatcher dispatcher:

        .. versionadded: 1.8.0

        """
        rs = request if isinstance(request, JSONRPC20BatchRequest) \
            else [request]
        responses = [r async for r in cls._get_responses(rs, dispatcher)
                     if r is not None]

        # notifications
        if not responses:
            return

        if isinstance(request, JSONRPC20BatchRequest):
            response = JSONRPC20BatchResponse(*responses)
            response.request = request
            return response
        else:
            return responses[0]

    @classmethod  # noqa: C901
    async def _get_responses(cls, requests, dispatcher):  # noqa: C901
        """ Response to each single JSON-RPC Request.

        :return iterator(JSONRPC20Response):

        .. versionadded: 1.9.0
          TypeError inside the function is distinguished from Invalid Params.

        """
        for request in requests:
            def make_response(**kwargs):
                response = cls.RESPONSE_CLASS_MAP[request.JSONRPC_VERSION](
                    _id=request._id, **kwargs)
                response.request = request
                return response

            output = None
            try:
                method = dispatcher[request.method]
            except KeyError:
                output = make_response(error=JSONRPCMethodNotFound()._data)
            else:
                try:
                    result = method(*request.args, **request.kwargs)
                    if inspect.isawaitable(result):
                        result = await result
                    # Find pydantic models and convert them to dict
                    if isinstance(result, pydantic.BaseModel):
                        result = result.dict()
                    elif isinstance(result, list):
                        result = [x.dict() if isinstance(x, pydantic.BaseModel) else x for x in result]
                except JSONRPCDispatchException as e:
                    output = make_response(error=e.error._data)
                except Exception as e:
                    err_args = e.args
                    if e.__class__.__name__ == 'ValidationError':
                        err_args = None
                    data = {
                        "type": e.__class__.__name__,
                        "args": err_args,
                        "message": str(e),
                    }

                    logger.exception("API Exception: {0}".format(data))

                    if isinstance(e, TypeError) and is_invalid_params(
                            method, *request.args, **request.kwargs):
                        output = make_response(
                            error=JSONRPCInvalidParams(data=data)._data)
                    else:
                        output = make_response(
                            error=JSONRPCServerError(data=data)._data)
                else:
                    output = make_response(result=result)
            finally:
                if not request.is_notification:
                    yield output



#    sizzlews
#    Copyright (C) 2020 Dmitry Berezovsky
#    The MIT License (MIT)
#    
#    Permission is hereby granted, free of charge, to any person obtaining
#    a copy of this software and associated documentation files
#    (the "Software"), to deal in the Software without restriction,
#    including without limitation the rights to use, copy, modify, merge,
#    publish, distribute, sublicense, and/or sell copies of the Software,
#    and to permit persons to whom the Software is furnished to do so,
#    subject to the following conditions:
#    
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#    
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
#    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
#    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import asyncio
from typing import List

import pydantic

from sizzlews.server.common import ClassBasedSizzleWSHandler
from sizzlews.server.tornado import bootstrap_torando_rpc_application


class MyDTO(pydantic.BaseModel):
    field1: int
    field2: str


class MyApi(ClassBasedSizzleWSHandler):
    METHOD_PREFXIX = "api."

    async def some_method(self, a: int, b):
        await asyncio.sleep(0.1)
        return a + b

    async def divide_by_zero(self, a: int):
        await asyncio.sleep(0.1)
        return 1 / 0

    async def my_dto_method(self):
        return MyDTO(field1=1, field2='str')

    async def returns_list_of_dtos(self) -> List[MyDTO]:
        return [MyDTO(field1=1, field2='str'), MyDTO(field1=2, field2='str2')]


if __name__ == "__main__":
    bootstrap_torando_rpc_application(MyApi(), url_path='/rpc')



#    sizzlews
#    Copyright (C) 2020 Dmitry Berezovsky
#    The MIT License (MIT)
#    
#    Permission is hereby granted, free of charge, to any person obtaining
#    a copy of this software and associated documentation files
#    (the "Software"), to deal in the Software without restriction,
#    including without limitation the rights to use, copy, modify, merge,
#    publish, distribute, sublicense, and/or sell copies of the Software,
#    and to permit persons to whom the Software is furnished to do so,
#    subject to the following conditions:
#    
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#    
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
#    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
#    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import datetime
import json
from abc import ABCMeta, abstractmethod
from json import JSONEncoder
from typing import Callable

from jsonrpc import Dispatcher
from jsonrpc.exceptions import JSONRPCInvalidRequest, JSONRPCInvalidRequestException, JSONRPCParseError
from jsonrpc.jsonrpc import JSONRPCRequest
from jsonrpc.jsonrpc2 import JSONRPC20Response

from sizzlews.jsonrpc.manager import JSONRPCResponseAsyncManager

API_METHOD_VERSION = 'version'
API_METHOD_PING = 'ping'

ANNOTATIONS_RPC_METHOD_PROPERTY = '_rpc_method'
ANNOTATIONS_IGNORE_PREFIX_PROPERTY = '_rpc_ignore_prefix'


class SizzleWSSession(object):

    def __init__(self, session_id: object):
        self.__id = session_id

    def id(self):
        return self.__id


class SizzleWSJsonEncoder(JSONEncoder):

    def default(self, o):
        if isinstance(o, datetime.datetime):
            return o.isoformat()


class SizzleWSHandler(object):

    def __init__(self, dispatcher: Dispatcher = None, expose_version_api=True, expose_ping_api=True) -> None:
        self.dispatcher = dispatcher or Dispatcher()
        self.__api_version = None  # type: str
        self.encoder = SizzleWSJsonEncoder()
        if expose_version_api:
            self.dispatcher[API_METHOD_VERSION] = self.m_get_version
        if expose_ping_api:
            self.dispatcher[API_METHOD_PING] = self.m_ping
        self.on_dispatcher_prepare(self.dispatcher)

    def on_dispatcher_prepare(self, dispatcher: Dispatcher):
        pass

    def m_get_version(self):
        return self.__api_version or "<unknown>"

    def m_ping(self):
        return "pong"

    def check_permissions(self):
        pass

    def _parse_request(self, request_str) -> JSONRPCRequest:
        if isinstance(request_str, bytes):
            request_str = request_str.decode("utf-8")
        data = json.loads(request_str)
        return JSONRPCRequest.from_data(data)

    def _build_response_str(self, response: JSONRPC20Response):
        return self.encoder.encode(response.data)

    async def handle(self, request_str, ctx: object) -> str:
        try:
            request = self._parse_request(request_str)
        except (TypeError, ValueError):
            return self._build_response_str(JSONRPC20Response(error=JSONRPCParseError()._data))
        except JSONRPCInvalidRequestException:
            return self._build_response_str(JSONRPC20Response(error=JSONRPCInvalidRequest()._data))

        # TODO: Middleware
        # TODO: Auth

        return self._build_response_str(await JSONRPCResponseAsyncManager.handle_request(request, self.dispatcher))


class BidirectionalSizzleWSSession(SizzleWSSession):

    def __init__(self, session_id: object, connection):
        super().__init__(session_id)
        self.connection = connection

    def id(self):
        return self.__id


class BiDirectionalSizzleWSHandler(SizzleWSHandler, metaclass=ABCMeta):

    def __init__(self, dispatcher: Dispatcher = None, expose_version_api=True, expose_ping_api=True) -> None:
        super().__init__(dispatcher, expose_version_api, expose_ping_api)
        self.sessions = dict()  # type: dict[int]

    @abstractmethod
    def post_message(self, message: JSONRPCRequest, session: BidirectionalSizzleWSSession):
        pass

    def register_session(self, session: SizzleWSSession):
        self.sessions[session.id] = session

    def broadcast(self, message: JSONRPCRequest, session_filter: Callable[[BidirectionalSizzleWSSession], bool]):
        for session_id, session in self.sessions.items():
            if session_filter is not None and session_filter(session):
                try:
                    self.post_message(message, session)
                except Exception:
                    # TODO: Log failure
                    pass


class MethodDiscoveryMixin(object):
    METHOD_PREFXIX = ""

    def on_dispatcher_prepare(self, dispatcher: Dispatcher):
        for name in dir(self):
            item = getattr(self, name)
            if callable(item) and hasattr(item, ANNOTATIONS_RPC_METHOD_PROPERTY):
                rpc_name = getattr(item, ANNOTATIONS_RPC_METHOD_PROPERTY)
                if not getattr(item, ANNOTATIONS_IGNORE_PREFIX_PROPERTY, False):
                    rpc_name = self.METHOD_PREFXIX + rpc_name
                dispatcher.add_method(item, rpc_name)


class ClassBasedSizzleWSHandler(SizzleWSHandler):
    METHOD_PREFXIX = ""

    def __init__(self, dispatcher: Dispatcher = None, expose_version_api=True, expose_ping_api=True) -> None:
        super().__init__(dispatcher, expose_version_api, expose_ping_api)

    def on_dispatcher_prepare(self, dispatcher: Dispatcher):
        dispatcher.build_method_map(self, self.METHOD_PREFXIX)



#    sizzlews
#    Copyright (C) 2020 Dmitry Berezovsky
#    The MIT License (MIT)
#    
#    Permission is hereby granted, free of charge, to any person obtaining
#    a copy of this software and associated documentation files
#    (the "Software"), to deal in the Software without restriction,
#    including without limitation the rights to use, copy, modify, merge,
#    publish, distribute, sublicense, and/or sell copies of the Software,
#    and to permit persons to whom the Software is furnished to do so,
#    subject to the following conditions:
#    
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#    
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
#    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
#    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from os import path
from setuptools import setup, find_packages

import sizzlews.__version__

src_dir = path.abspath(path.dirname(__file__))
root_dir = path.join(src_dir, '..')

version = sizzlews.__version__.__version__

readme_file = path.join(root_dir, 'README.md')

try:
    from m2r import parse_from_file

    long_description = parse_from_file(readme_file)
except ImportError:
    # m2r may not be installed in user environment
    with open(readme_file) as f:
        long_description = f.read()

setup(
    name='ws-sizzle',
    # Semantic versioning should be used:
    # https://packaging.python.org/distributing/?highlight=entry_points#semantic-versioning-preferred
    version=version,
    description='Lightweight library for streamlining JSON RPC client and server creating. '
                'Supports WebSockets and HTTP transport',
    long_description=long_description,
    url='https://github.com/corvis/ws-sizzle',
    keywords='json, json-rpc, rpc, jsonrpc, jsonrpc-websockets, bidirectional-jsonrpc',
    # Author
    author='Dmitry Berezovsky',

    # License
    license='MIT',

    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Internet :: WWW/HTTP',

        # License (should match "license" above)
        'License :: OSI Approved :: MIT License',
        # Python versions support
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
    ],

    # Structure
    packages=find_packages(include=['sizzlews', 'sizzlews.*']),

    # Mandatory requirements
    install_requires=[
        'json-rpc>=1.13.0',
        'pydantic>=1.6',
        # 'typing>=3.6',
    ],
    # Extra dependencies might be installed with:
    # pip install ws-sizzle[tornado,websockets,etc]
    extras_require={
        'tornado': ['tornado'],
        'websockets': ['websockets'],
        'requests-client': ['requests'],
        'aiohttp': ['aiohttp']
    },

)



#    sizzlews
#    Copyright (C) 2020 Dmitry Berezovsky
#    The MIT License (MIT)
#    
#    Permission is hereby granted, free of charge, to any person obtaining
#    a copy of this software and associated documentation files
#    (the "Software"), to deal in the Software without restriction,
#    including without limitation the rights to use, copy, modify, merge,
#    publish, distribute, sublicense, and/or sell copies of the Software,
#    and to permit persons to whom the Software is furnished to do so,
#    subject to the following conditions:
#    
#    The above copyright notice and this permission notice shall be
#    included in all copies or substantial portions of the Software.
#    
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
#    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
#    CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
#    TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
#    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import asyncio
from typing import Any

import tornado.ioloop
import tornado.web
from tornado import httputil

from sizzlews.server.common import SizzleWSHandler


class TornadoHttpSizzleWSHandler(tornado.web.RequestHandler):

    def __init__(self, application: "tornado.web.Application", request: httputil.HTTPServerRequest,
                 **kwargs: Any) -> None:
        super().__init__(application, request, **kwargs)
        # self._api_handler = None  # type: SizzleWSApiHandler

    def initialize(self, api_handler: SizzleWSHandler):
        self._api_handler = api_handler

    def set_default_headers(self):
        self.set_header("Content-Type", 'application/json')

    async def post(self):
        self.write(await self._api_handler.handle(self.request.body, str(self.request.body)))


def bootstrap_torando_rpc_application(api_handler: SizzleWSHandler, port=8888, url_path='/'):
    app = tornado.web.Application([
        (url_path, TornadoHttpSizzleWSHandler, dict(api_handler=api_handler))
    ])
    app.listen(port)
    if not asyncio.get_event_loop().is_running():
        tornado.ioloop.IOLoop.current().start()
    return app
